class NetboxAllocationError(Exception):
    """
    Raised when there is an error in allocating resource in Netbox
    """


class UnsupportedNetboxVersion(Exception):
    """
    Raised when there is an error in allocating resource in Netbox
    """
